from threading import Timer


class BulkQueue():

    def __init__(self, es_client, threshold=10, flush_time=30):
        self.es_client = es_client
        self.threshold = threshold
        self.list = list()
        self.flush_time = flush_time
        self.time_thread = Timer(self.flush_time, self.timer_over)
        self.running = False

    def timer_over(self):
        self.flush()
        self.time_thread = Timer(self.flush_time, self.timer_over)
        self.start_timer()

    def start_timer(self):
        """ Start timer thread that flushes queue every X seconds """
        self.time_thread.start()
        self.running = True

    def stop_timer(self):
        """ Stop timer thread """
        self.time_thread.cancel()
        self.running = False

    def started(self):
        return self.running

    def add(self, index, doc_type, body, doc_id=None):
        """ Add event to queue, flushing if we hit the threshold """
        bulk_doc = {
            "_index": index,
            "_type": doc_type,
            "_id": doc_id,
            "_source": body
        }
        self.list.append(bulk_doc)
        if self.size() >= self.threshold:
            self.flush()

    def size(self):
        """ Size of the queue structure """
        return len(self.list)

    def flush(self):
        """ Write all stored events to ES """
        self.es_client.save_documents(self.list)
        self.list = list()



import os
import pynsive
from operator import itemgetter
from utilities.dict2List import dict2List
from utilities.logger import logger


class PluginSet(object):
    def __init__(self, plugin_location, enabled_plugins=None):
        self.plugin_location = plugin_location
        self.enabled_plugins = self.identify_plugins(enabled_plugins)

    def identify_plugins(self, enabled_plugins):
        if not os.path.exists(self.plugin_location):
            return []

        module_name = os.path.basename(self.plugin_location)
        root_plugin_directory = os.path.join(self.plugin_location, '..')

        plugin_manager = pynsive.PluginManager()
        plugin_manager.plug_into(root_plugin_directory)

        plugins = []

        found_modules = pynsive.list_modules(module_name)
        for found_module in found_modules:
            module_filename, module_name = found_module.split('.')
            if enabled_plugins is not None and module_name not in enabled_plugins:
                # Skip this plugin since it's not listed as enabled plugins
                # as long as we have specified some enabled plugins though
                # this allows us to specify no specific plugins and get all of them
                continue

            module_obj = pynsive.import_module(found_module)
            reload(module_obj)
            plugin_class_obj = module_obj.message()

            if 'priority' in dir(plugin_class_obj):
                priority = plugin_class_obj.priority
            else:
                priority = 100

            logger.info('[*] plugin {0} registered to receive messages with {1}'.format(module_name, plugin_class_obj.registration))
            plugins.append(
                {
                    'plugin_class': plugin_class_obj,
                    'registration': plugin_class_obj.registration,
                    'priority': priority
                }
            )
        return plugins

    @property
    def ordered_enabled_plugins(self):
        return sorted(self.enabled_plugins, key=itemgetter('priority'), reverse=False)

    def run_plugins(self, message, metadata=None):
        '''compare the message to the plugin registrations.
           plugins register with a list of keys or values
           or values they want to match on
           this function compares that registration list
           to the current message and sends the message to plugins
           in order
        '''
        if not isinstance(message, dict):
            raise TypeError('event is type {0}, should be a dict'.format(type(message)))

        for plugin in self.ordered_enabled_plugins:
            send = False
            message_fields = [e for e in dict2List(message)]
            # this is to make it so we can match on all fields
            message_fields.append('*')
            if isinstance(plugin['registration'], list):
                if set(plugin['registration']).intersection(message_fields):
                    send = True
            elif isinstance(plugin['registration'], str):
                if plugin['registration'] in message_fields:
                    send = True
            if send:
                try:
                    (message, metadata) = self.send_message_to_plugin(plugin_class=plugin['plugin_class'], message=message, metadata=metadata)
                except Exception as e:
                    logger.error('Received exception in {0}: message: {1}\n{2}'.format(plugin['plugin_class'], message, e.message))
                if message is None:
                    return (message, metadata)
        return (message, metadata)

    def send_message_to_plugin(self, plugin_class, message, metadata=None):
        '''moving this logic to a separate function allows
           different types of plugin_sets, such as alerts that might not care
           about receiving metadata in its plugins
        '''
        return plugin_class.onMessage(message, metadata)



#!/usr/bin/env python

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# Copyright (c) 2017 Mozilla Corporation


def SimpleResults(input_results):
    converted_results = {
        'meta': {
            'timed_out': input_results.timed_out,
        },
        'hits': []
    }
    for hit in input_results.hits:
        hit_dict = {
            '_id': hit.meta.id,
            '_type': hit.meta.doc_type,
            '_index': hit.meta.index,
            '_score': hit.meta.score,
            '_source': hit.to_dict()
        }

        converted_results['hits'].append(hit_dict)

    return converted_results



#!/usr/bin/env python

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# Copyright (c) 2017 Mozilla Corporation


from mozdef_util.utilities.toUTC import toUTC

from datetime import datetime
from datetime import timedelta

from range_match import RangeMatch
from boolean_match import BooleanMatch


class SearchQuery(object):
    def __init__(self, *args, **kwargs):
        self.date_timedelta = dict(kwargs)
        self.must = []
        self.must_not = []
        self.should = []
        self.aggregation = []

    def append_to_array(self, in_array, in_obj):
        """
        Allow a list or a specific filter/query object to
        get added to build a query
        """
        if isinstance(in_obj, list):
            for key in in_obj:
                in_array.append(key)
        else:
            in_array.append(in_obj)

    def add_must(self, input_obj):
        self.append_to_array(self.must, input_obj)

    def add_must_not(self, input_obj):
        self.append_to_array(self.must_not, input_obj)

    def add_should(self, input_obj):
        self.append_to_array(self.should, input_obj)

    def add_aggregation(self, input_obj):
        self.append_to_array(self.aggregation, input_obj)

    def execute(self, elasticsearch_client, indices=['events', 'events-previous'], size=1000, request_timeout=30):
        if self.must == [] and self.must_not == [] and self.should == [] and self.aggregation == []:
            raise AttributeError('Must define a must, must_not, should query, or aggregation')

        if self.date_timedelta:
            end_date = toUTC(datetime.now())
            begin_date = toUTC(datetime.now() - timedelta(**self.date_timedelta))
            utc_range_query = RangeMatch('utctimestamp', begin_date, end_date)
            received_range_query = RangeMatch('receivedtimestamp', begin_date, end_date)
            range_query = utc_range_query | received_range_query
            self.add_must(range_query)

        search_query = None
        search_query = BooleanMatch(must=self.must, must_not=self.must_not, should=self.should)

        results = []
        if len(self.aggregation) == 0:
            results = elasticsearch_client.search(search_query, indices, size, request_timeout)
        else:
            results = elasticsearch_client.aggregated_search(search_query, indices, self.aggregation, size, request_timeout)

        return results



import os
import geoip2.database


class GeoIP(object):
    def __init__(self, db_location=None):
        if db_location is None:
            db_location = os.path.join(os.path.dirname(os.path.abspath(__file__)), "../data/GeoLite2-City.mmdb")
        try:
            self.db = geoip2.database.Reader(db_location)
        except IOError:
            self.error = 'No Geolite DB Found!'

    def lookup_ip(self, ip):
        if hasattr(self, 'error'):
            return {'error': self.error}

        try:
            result = self.db.city(ip)
        except Exception as e:
            return {'error': e.message}

        geo_dict = {}
        geo_dict['city'] = result.city.name
        geo_dict['continent'] = result.continent.code
        geo_dict['country_code'] = result.country.iso_code
        geo_dict['country_name'] = result.country.name
        geo_dict['dma_code'] = result.location.metro_code
        geo_dict['latitude'] = result.location.latitude
        geo_dict['longitude'] = result.location.longitude
        geo_dict['metro_code'] = ""
        if result.city.names:
            geo_dict['metro_code'] = result.city.names['en']
        geo_dict['postal_code'] = result.postal.code
        geo_dict['region_code'] = ""
        if result.subdivisions:
            geo_dict['region_code'] = result.subdivisions[0].iso_code
            geo_dict['metro_code'] += ', ' + result.subdivisions[0].iso_code
        geo_dict['time_zone'] = result.location.time_zone

        return geo_dict



import json

from elasticsearch import Elasticsearch
from elasticsearch_dsl import Search
from elasticsearch.exceptions import NotFoundError
from elasticsearch.helpers import bulk, BulkIndexError

from query_models import SearchQuery, TermMatch, AggregatedResults, SimpleResults
from bulk_queue import BulkQueue

from utilities.logger import logger

from event import Event


class ElasticsearchBadServer(Exception):
    def __str__(self):
        return "Bad ES Server defined"


class ElasticsearchException(Exception):
    def __str__(self):
        return "Exception in ES encountered"


class ElasticsearchInvalidIndex(Exception):
    def __init__(self, index_name):
        self.index_name = index_name

    def __str__(self):
        return "Invalid index: " + str(self.index_name)


class ElasticsearchClient():

    def __init__(self, servers, bulk_amount=100, bulk_refresh_time=30):
        self.es_connection = Elasticsearch(servers)
        self.es_connection.ping()
        self.bulk_queue = BulkQueue(self, threshold=bulk_amount, flush_time=bulk_refresh_time)

    def delete_index(self, index_name, ignore_fail=False):
        ignore_codes = []
        if ignore_fail is True:
            ignore_codes = [400, 404]

        self.es_connection.indices.delete(index=index_name, ignore=ignore_codes)

    def get_indices(self):
        return self.es_connection.indices.stats()['indices'].keys()

    def index_exists(self, index_name):
        return self.es_connection.indices.exists(index_name)

    def create_index(self, index_name, index_config=None):
        if not index_config:
            index_config = '''
            {
              "mappings":{}
            }'''
        self.es_connection.indices.create(index=index_name, update_all_types='true', body=index_config)

    def create_alias(self, alias, index):
        actions = []
        if self.es_connection.indices.exists_alias('*', alias):
            actions.append({
                'remove': {'index': '*', 'alias': alias}
            })

        actions.append({
            'add': {'index': index, 'alias': alias}
        })
        self.es_connection.indices.update_aliases(dict(actions=actions))

    def create_alias_multiple_indices(self, alias_name, indices):
        actions = []
        if self.es_connection.indices.exists_alias('*', alias_name):
            actions.append({
                'remove': {'index': '*', 'alias': alias_name}
            })

        for index in indices:
            actions.append({'add': {'index': index, 'alias': alias_name}})
        self.es_connection.indices.update_aliases(dict(actions=actions))

    def get_alias(self, alias_name):
        return self.es_connection.indices.get_alias(index='*', name=alias_name).keys()

    def flush(self, index_name):
        self.es_connection.indices.flush(index=index_name)

    def search(self, search_query, indices, size, request_timeout):
        results = []
        try:
            results = Search(using=self.es_connection, index=indices).params(size=size, request_timeout=request_timeout).filter(search_query).execute()
        except NotFoundError:
            raise ElasticsearchInvalidIndex(indices)

        result_set = SimpleResults(results)
        return result_set

    def aggregated_search(self, search_query, indices, aggregations, size, request_timeout):
        search_obj = Search(using=self.es_connection, index=indices).params(size=size, request_timeout=request_timeout)
        query_obj = search_obj.filter(search_query)
        for aggregation in aggregations:
            query_obj.aggs.bucket(name=aggregation.to_dict()['terms']['field'], agg_type=aggregation)
        results = query_obj.execute()

        result_set = AggregatedResults(results)
        return result_set

    def save_documents(self, documents):
        try:
            bulk(self.es_connection, documents)
        except BulkIndexError as e:
            logger.error("Error bulk indexing: " + str(e))

    def start_bulk_timer(self):
        if not self.bulk_queue.started():
            self.bulk_queue.start_timer()

    def finish_bulk(self):
        self.bulk_queue.stop_timer()

    def __bulk_save_document(self, index, doc_type, body, doc_id=None):
        self.start_bulk_timer()
        self.bulk_queue.add(index=index, doc_type=doc_type, body=body, doc_id=doc_id)

    def __save_document(self, index, doc_type, body, doc_id=None, bulk=False):
        if bulk:
            self.__bulk_save_document(index=index, doc_type=doc_type, body=body, doc_id=doc_id)
        else:
            return self.es_connection.index(index=index, doc_type=doc_type, id=doc_id, body=body)

    def __parse_document(self, body, doc_type):
        if type(body) is str:
            body = json.loads(body)

        if '_type' in body:
            doc_type = body['_type']

        doc_body = body
        if '_source' in body:
            doc_body = body['_source']
        return doc_body, doc_type

    def save_object(self, body, index, doc_type, doc_id=None, bulk=False):
        doc_body, doc_type = self.__parse_document(body, doc_type)
        return self.__save_document(index=index, doc_type=doc_type, body=doc_body, doc_id=doc_id, bulk=bulk)

    def save_alert(self, body, index='alerts', doc_type='alert', doc_id=None, bulk=False):
        doc_body, doc_type = self.__parse_document(body, doc_type)
        return self.__save_document(index=index, doc_type=doc_type, body=doc_body, doc_id=doc_id, bulk=bulk)

    def save_event(self, body, index='events', doc_type='event', doc_id=None, bulk=False):
        doc_body, doc_type = self.__parse_document(body, doc_type)
        event = Event(doc_body)
        event.add_required_fields()
        return self.__save_document(index=index, doc_type=doc_type, body=event, doc_id=doc_id, bulk=bulk)

    def get_object_by_id(self, object_id, indices):
        id_match = TermMatch('_id', object_id)
        search_query = SearchQuery()
        search_query.add_must(id_match)
        results = search_query.execute(self, indices=indices)
        if len(results['hits']) == 0:
            return None
        else:
            return results['hits'][0]

    def get_alert_by_id(self, alert_id):
        return self.get_object_by_id(alert_id, ['alerts'])

    def get_event_by_id(self, event_id):
        return self.get_object_by_id(event_id, ['events'])

    def save_dashboard(self, dash_file, dash_name):
        f = open(dash_file)
        dashboardjson = json.load(f)
        f.close()
        title = dashboardjson['title']
        dashid = dash_name.replace(' ', '-')
        if dash_name:
            title = dash_name
        dashboarddata = {
            "user": "guest",
            "group": "guest",
            "title": title,
            "dashboard": json.dumps(dashboardjson)
        }

        return self.es_connection.index(index='.kibana', doc_type='dashboard', body=dashboarddata, id=dashid)

    def get_cluster_health(self):
        health_dict = self.es_connection.cluster.health()
        # To line up with the health stats from ES1, we're
        # removing certain keys
        health_dict.pop('active_shards_percent_as_number', None)
        health_dict.pop('delayed_unassigned_shards', None)
        health_dict.pop('number_of_in_flight_fetch', None)
        health_dict.pop('number_of_pending_tasks', None)
        health_dict.pop('task_max_waiting_in_queue_millis', None)

        return health_dict



import json


class StateParsingError(ValueError):
    pass


class State:
    def __init__(self, filename):
        '''Set the filename and populate self.data by calling self.read_stat_file()'''
        self.filename = filename
        self.read()

    def read(self):
        '''Populate self.data by reading and parsing the state file'''
        try:
            with open(self.filename, 'r') as f:
                self.data = json.load(f)
        except IOError:
            self.data = {}
        except ValueError:
            raise StateParsingError(
                "%s state file found but isn't a recognized json format" % self.filename)

    def save(self):
        '''Write the self.data value into the state file'''
        with open(self.filename, 'w') as f:
            json.dump(self.data, f, sort_keys=True, indent=4, separators=(',', ': '))



#!/usr/bin/env python

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# Copyright (c) 2017 Mozilla Corporation


class DotDict(dict):
    '''dict.item notation for dict()'s'''
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __init__(self, dct={}):
        for key, value in dct.items():
            if hasattr(value, 'keys'):
                value = DotDict(value)
            self[key] = value

    def get(self, key):
        """get to allow for dot string notation
        :param str key: Key in dot-notation (e.g. 'foo.lol').
        :return: value. None if no value was found.
        """
        try:
            return self.__lookup(self, key)
        except KeyError:
            return None

    def __lookup(self, dct, key):
        """Checks dct recursive to find the value for key.
        Is used by get() interanlly.
        :param dict dct: input dictionary
        :param str key: The key we are looking for.
        :return: The  value.
        :raise KeyError: If the given key is not found
        """
        if '.' in key:
            key, node = key.split('.', 1)
            return self.__lookup(dct[key], node)
        else:
            return dct[key]



#!/usr/bin/env python

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# Copyright (c) 2017 Mozilla Corporation

import logging
import sys
from datetime import datetime
from logging.handlers import SysLogHandler

from toUTC import toUTC


def loggerTimeStamp(self, record, datefmt=None):
    return toUTC(datetime.now()).isoformat()


def initLogger(options=None):
    logger.level = logging.INFO
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    formatter.formatTime = loggerTimeStamp

    output = ''
    try:
        output = options.output
    except Exception:
        output = 'stderr'

    for handler in logger.handlers:
        logger.removeHandler(handler)

    if output == 'syslog':
        logger.addHandler(SysLogHandler(address=(options.sysloghostname, options.syslogport)))
    else:
        sh = logging.StreamHandler(sys.stderr)
        sh.setFormatter(formatter)
        logger.addHandler(sh)


logger = logging.getLogger(sys.argv[0])
initLogger()



from datetime import datetime
import socket

from utilities.toUTC import toUTC


class Event(dict):

    # We set default vaules so that we can later
    # create an alert around these, and know when events
    # have to use defaults
    DEFAULT_STRING = 'UNKNOWN'

    def add_required_fields(self):
        if 'receivedtimestamp' not in self:
            self['receivedtimestamp'] = toUTC(datetime.now()).isoformat()
        if 'utctimestamp' not in self:
            self['utctimestamp'] = toUTC(datetime.now()).isoformat()
        if 'timestamp' not in self:
            self['timestamp'] = toUTC(datetime.now()).isoformat()
        if 'mozdefhostname' not in self:
            self['mozdefhostname'] = socket.gethostname()
        if 'tags' not in self:
            self['tags'] = []
        if 'category' not in self:
            self['category'] = self.DEFAULT_STRING
        if 'hostname' not in self:
            self['hostname'] = self.DEFAULT_STRING
        if 'processid' not in self:
            self['processid'] = self.DEFAULT_STRING
        if 'processname' not in self:
            self['processname'] = self.DEFAULT_STRING
        if 'severity' not in self:
            self['severity'] = self.DEFAULT_STRING
        if 'source' not in self:
            self['source'] = self.DEFAULT_STRING
        if 'summary' not in self:
            self['summary'] = self.DEFAULT_STRING
        if 'details' not in self:
            self['details'] = {}




def key_exists(search_key, inputed_dict):
    '''
    Given a search key which is dot notated
    return wether or not that key exists in a dictionary
    '''
    num_levels = search_key.split(".")
    if len(num_levels) == 0:
        return False
    current_pointer = inputed_dict
    for updated_key in num_levels:
        if current_pointer is None:
            return False
        if updated_key == num_levels[-1]:
            return updated_key in current_pointer
        if updated_key in current_pointer:
            current_pointer = current_pointer[updated_key]
        else:
            return False
